/*
 * A module to map remote access to /proc to local access 
 * interfaces in this directory.
 *
 * October, 1998 TiVo, Inc. James M. Barton
 * This code is released to the public domain. Do with it as you will.
 * There is no warranty, express or implied, that this code is fit for
 * any purpose whatsoever.
 */

/*
 * Remote access protocol:
 *
 * On accepting the connection, access to files in /proc (only)
 * is permitted, but full pathnames are expected. The protocol
 * is strictly request/response, delineated bu newlines.
 *
 * Q: SCAN dir
 * A: ACK, NACK
 * Resets the directory scanner to the first file in the directory
 * specified.
 *
 * Q: NEXT
 * A: filename, NULL
 * Returns next file in the scanned directory, or an empty line
 *
 * Q: READ file
 * A: nbytes
 * ... nbytes of data ...
 * Read the file named and pass the contents back. A -1 count
 * means no such file.
 */

#include    <sys/types.h>
#include    <sys/socket.h>
#include    <netinet/in.h>
#include    <netdb.h>
#include    <string.h>
#include    <stdlib.h>
#include    <stdio.h>
#include    <unistd.h>
#include    <signal.h>
#include    <assert.h>
#include    "psdata.h"
#include    <sys/stat.h>
#include    <fcntl.h>

#define     TARGET_NODE     "PROCINFO_TARGET"
#define     TARGET_DEFPORT  25007
#undef      pinfoDEBUG

static int  sfd = -1;
static FILE *fsfd;
static char sbuf[BUFSIZ];
static int  _proc_remote = 1;

static inline void
die() {
    fprintf( stderr, "fatal error reading from remote procd\n" );
    kill(getpid(), SIGINT);
}

static int
_proc_gets() {
    char *p = sbuf;
    while ((p - sbuf < BUFSIZ) && (read(sfd, p, 1) == 1)) {
        if (*p == '\n') {
            p++;
            *p = '\0';
            return p - sbuf;
        }
        p++;
    }
    return -1;
}

int
_proc_readsock( int sock, char *buf, int len ) {
    int sofar = 0;
    while ( sofar < len ) {
        int rval = read( sock, buf, len - sofar );
        if ( rval <= 0 )
            die();
        sofar += rval;
        buf += rval;
    }
    return len;
}

static int
_proc_mkaddr(const char *pAddr, struct sockaddr *s, int *alen) {
    struct sockaddr_in *sin = (struct sockaddr_in *) s;
    char *node = pAddr;
    int nodelen;
    int port;
    struct hostent *hep;

    while (*node != '\0' && *node != ':') node++;
    nodelen = node - pAddr;
    if (*node == ':') {
        char *sport = node + 1;
        port = strtol(sport, 0, 0);
        node = (char *) alloca(nodelen + 1);
        strncpy(node, pAddr, nodelen);
        node[nodelen] = '\0';
    } else {
        node = pAddr;
        port = TARGET_DEFPORT;
    }

    if ((hep = gethostbyname(node)) == 0)
        return 0;
    sin->sin_addr = *((struct in_addr *) hep->h_addr);
    sin->sin_port = htons(port);
    sin->sin_family = AF_INET;
    *alen = sizeof(struct sockaddr_in);
    return 1;
}

int
_proc_open() {
    if (!_proc_remote)
        return 0;
    if (sfd == -1) {
        char *ps;
        struct sockaddr s;
        int addrlen;

        if ((ps = getenv(TARGET_NODE)) == 0) {
            _proc_remote = 0;
            return 0;
        }
        if ((sfd = socket(AF_INET, SOCK_STREAM, 6)) == -1) {
            perror("Can't create socket");
            die();
        }
        if (!_proc_mkaddr(ps, &s, &addrlen)) {
            fprintf(stderr, "Can't resolve %s to anything useful\n", ps);
            die();
        }
        if (connect(sfd, &s, addrlen) == -1) {
            perror("Can't connect");
            die();
        }
        fsfd = fdopen(sfd, "w");
        setvbuf(fsfd, 0, _IONBF, 0);
    }
    return 1;
}

char *
_proc_f2buf(char *pnode, char *pbuf, int *plen) {
    int blen;
    if (!_proc_open())
        return 0;
    fprintf(fsfd, "READ %s\n", pnode);
    if (_proc_gets() <= 0)
        die();
    if (sscanf(sbuf, "%d\n", &blen) != 1) {
        fprintf(stderr, "Unexpected protocol error 1\n");
        return 0;
    }
    if (blen < 0)
        return 0;
    if (pbuf) {
        if (blen < *plen) {
            _proc_readsock( sfd, pbuf, blen );
            pbuf[blen] = '\0';
            *plen = blen + 1;
        }
        else {
            char throwbuf[64];
            int leftover = blen - *plen - 1;
            _proc_readsock( sfd, pbuf, *plen - 1 );
            pbuf[*plen] = '\0';
            while (leftover > 0) {
                int n = (leftover > 64 ? 64 : leftover);
                _proc_readsock( sfd, throwbuf, n );
                leftover -= n;
            }
        }
    }
    else {
        pbuf = xmalloc(blen + 1);
        _proc_readsock( sfd, pbuf, blen );
        pbuf[blen] = '\0';
        *plen = blen + 1;
    }
    return pbuf;
}

int
_proc_scandir(char *dir) {
    if (!_proc_open())
        return 0;
    fprintf(fsfd, "SCAN %s\n", dir);
    if (_proc_gets() <= 0)
        die();
    if (strcmp(sbuf, "ACK\n"))
        return 0;
    return 1;
}

char *
_proc_nextdir() {
    if (!_proc_open())
        return 0;
    fprintf(fsfd, "NEXT\n");
    if (_proc_gets() <= 0)
        die();
    sbuf[strlen(sbuf)-1] = '\0';
    return sbuf;
}

int
_proc_kill( int pid, int sig ) {
    if (!_proc_open())
        return 0;
    fprintf( fsfd, "KILL %d %d\n", pid, sig );
    if ( _proc_gets() <= 0 )
        die();
    return strtol(sbuf, 0, 0);
}

int
_proc_setpri( int cmd, int pid, int val ) {
    if (!_proc_open())
        return 0;
    fprintf( fsfd, "SETP %d %d %d\n", cmd, pid, val );
    if ( _proc_gets() <= 0 )
        die();
    return strtol(sbuf, 0, 0);
}

int
_proc_gdb( int pid ) {
    if (!_proc_open())
        return 0;
    fprintf( fsfd, "GDBS %d\n", pid );
    if ( _proc_gets() <= 0)
        die();
    return strtol(sbuf, 0, 0);
}

int
_proc_gprof( int pid, int mode ) {
    if (!_proc_open())
        return 0;
    fprintf( fsfd, "GPRF %d %d\n", pid, mode );
    if ( _proc_gets() <= 0)
        die();
    return strtol(sbuf, 0, 0);
}

void
start_gdbx( int pid, int port, char *objfile ) {
	char tmpfile[80];
	char tmpstring[80];
	char *ps;
	FILE *f;
	char *debugger;
	char cmdline[256];
	char errmsg[128];
	/*extern showerr(char *);*/

        if ((ps = getenv(TARGET_NODE)) == 0) {
		sprintf(errmsg,"No remote node to connect to\n");
		/*showerr(errmsg);*/
		return;
	}
	sprintf(tmpfile, "/tmp/gdbx.%d", port);
	if ((f = fopen(tmpfile, "w")) < 0) {
		sprintf(errmsg, "Can't open tmp file: %s.\n", tmpfile);
		/*showerr(errmsg);*/
		return;
	}
	if (debugger = getenv("TOP_USEDDD")) {
		sprintf(cmdline, "/usr/local/gcc-ppc/bin/ddd --debugger /usr/local/gcc-ppc/bin/gdb -x %s %s", 
		tmpfile, objfile);
	} else {
		sprintf(cmdline, "xterm -title %s:%d -n %d -e /usr/local/gcc-ppc/bin/gdb -x %s %s",
		ps, pid, pid, tmpfile, objfile);
	}
	sprintf(tmpstring, "target remote %s:%d\n", ps, port);
	fwrite(tmpstring, strlen(tmpstring), 1, f);
	fclose(f);
	if (!fork()) {
		execl("/bin/sh", "sh", "-c", cmdline, (char *)0);
		sprintf(errmsg, "Can't exec %s\n", cmdline);
		/*showerr(errmsg);*/
		exit(0);
	}
	return;
}

int
_proc_setrtpri( int pid, int policy, int pri ) {
    if (!_proc_open())
        return 0;
    fprintf( fsfd, "RTPR %d %d %d\n", pid, policy, pri );
    if ( _proc_gets() <= 0 )
        die();
    return strtol(sbuf, 0, 0);
}

struct {
    int grabbing;
    char *g_name;
    int g_pid;
    char *g_stat;
    int g_stat_len;
    char *g_statm;
    int g_statm_len;
    char *g_cmdline;
    int g_cmdline_len;
    char *g_environ;
    int g_environ_len;
} grabinfo = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

int
_proc_begingrab( char *which ) {
    assert( !grabinfo.grabbing );
    fprintf( fsfd, "GRAB %s\n", which );
    grabinfo.grabbing = 1;
#ifdef pinfoDEBUG
    printf( "Entering grab mode\n" );
#endif
    return 1;
}

int
_proc_grabnext() {
    int rval;
    int len;

    if ( grabinfo.g_name ) free( grabinfo.g_name ), grabinfo.g_name = 0;
    if ( grabinfo.g_stat ) free( grabinfo.g_stat ), grabinfo.g_stat = 0;
    if ( grabinfo.g_statm ) free( grabinfo.g_statm ), grabinfo.g_statm = 0;
    if ( grabinfo.g_cmdline ) free( grabinfo.g_cmdline ), grabinfo.g_cmdline = 0;
    if ( grabinfo.g_environ ) free( grabinfo.g_environ ), grabinfo.g_environ = 0;

    while ( ( rval = _proc_gets() ) > 0 ) {
#ifdef pinfoDEBUG
        printf( "%s", sbuf );
#endif
        switch ( sbuf[0] ) {
        case 'P':
            grabinfo.g_name = (char *) xmalloc( strlen(&sbuf[2]) + 1 );
            strcpy( grabinfo.g_name, &sbuf[2] );
            grabinfo.g_pid = strtol( &sbuf[2], 0, 0 );
#ifdef pinfoDEBUG
            printf( "Grabbed %d\n", grabinfo.g_pid );
#endif
            break;
        case 'G':       /* grap operation complete */
            grabinfo.grabbing = 0;
#ifdef pinfoDEBUG
            printf( "Grab complete\n" );
#endif
            return 0;
        case 'Z':       /* record complete */
#ifdef pinfoDEBUG
            printf( "Record complete\n" );
#endif
            return 1;
        case 'S':
            sscanf( &sbuf[2], "%d", &len );
            if ( len ) {
                grabinfo.g_stat = (char *) xmalloc( len + 1 );
                grabinfo.g_stat_len = len + 1;
                grabinfo.g_stat[len] = '\0';
                _proc_readsock( sfd, grabinfo.g_stat, len );
#ifdef pinfoDEBUG
                printf( "(read %d bytes of data)\n", len );
#endif
            }
            break;
        case 'M':
            sscanf( &sbuf[2], "%d", &len );
            if ( len ) {
                grabinfo.g_statm = (char *) xmalloc( len + 1 );
                grabinfo.g_statm_len = len + 1;
                grabinfo.g_statm[len] = '\0';
                _proc_readsock( sfd, grabinfo.g_statm, len );
#ifdef pinfoDEBUG
                printf( "(read %d bytes of data)\n", len );
#endif
            }
            break;
        case 'C':
            sscanf( &sbuf[2], "%d", &len );
            if ( len ) {
                grabinfo.g_cmdline = (char *) xmalloc( len + 1 );
                grabinfo.g_cmdline_len = len + 1;
                grabinfo.g_cmdline[len] = '\0';
                _proc_readsock( sfd, grabinfo.g_cmdline, len );
#ifdef pinfoDEBUG
                printf( "(read %d bytes of data)\n", len );
#endif
            }
            break;
        case 'E':
            sscanf( &sbuf[2], "%d", &len );
            if ( len ) {
                grabinfo.g_environ = (char *) xmalloc( len + 1 );
                grabinfo.g_environ_len = len + 1;
                grabinfo.g_environ[len] = '\0';
                _proc_readsock( sfd, grabinfo.g_environ, len );
#ifdef pinfoDEBUG
                printf( "(read %d bytes of data)\n", len );
#endif
            }
            break;
        }
    }
    die();
    return 0;   /* NOTREACHED */
}

void
_proc_endgrab() {
    if ( !grabinfo.grabbing )
        return;
    while ( _proc_grabnext() );
}

int
_proc_get_pid() {
    if ( !grabinfo.grabbing )
        return 0;
    return grabinfo.g_pid;
}

char *
_proc_get_stat() {
    if ( !grabinfo.grabbing )
        return 0;
    if ( grabinfo.g_stat )
        return grabinfo.g_stat;
    else
        return "";
}

char *
_proc_get_statm() {
    if ( !grabinfo.grabbing )
        return 0;
    if ( grabinfo.g_statm )
        return grabinfo.g_statm;
    else
        return "";
}

char *
_proc_get_cmdline( int *len ) {
    char *tmp = grabinfo.g_cmdline;
    if ( !grabinfo.grabbing )
        return 0;
    *len = grabinfo.g_cmdline_len;
    grabinfo.g_cmdline = 0;
    return tmp;
}

char *
_proc_get_environ( int *len ) {
    char *tmp = grabinfo.g_environ;
    if ( !grabinfo.grabbing )
        return 0;
    *len = grabinfo.g_environ_len;
    grabinfo.g_environ = 0;
    return tmp;
}
